from types import ModuleType
from typing import ForwardRef

from lazi.conf import conf
from lazi.util import debug

__all__ = "Module",

Spec = ForwardRef("Spec")
Loader = ForwardRef("Loader")

# See https://peps.python.org/pep-0451/#attributes
MODULE_SPEC_ATTR_MAP = dict(
    __name__="name",
    __loader__="loader",
    __package__="parent",
    __file__="origin",
    __cached__="cached",
    __path__="submodule_search_locations",
)

GETATTR_PASS = ("__spec__", "__class__") + (("__dict__",) if not conf.NO_DICT_LAZY_ATTR else ())
SETATTR_PASS = GETATTR_PASS + tuple(MODULE_SPEC_ATTR_MAP)


class Module(ModuleType):
    __name__: str
    __loader__: Loader
    __package__: str
    __spec__: Spec

    def __init__(self, spec: Spec, module: ModuleType):
        super().__init__(spec.name)
        self.__spec__ = module.__spec__ = spec
        spec.target = module

    def __getattribute__(self, attr):
        spec = super().__getattribute__("__spec__")

        assert None is debug.traced(3, f"[{id(self)}] <GET> {spec.loader_state} {spec.name}[.{attr}]")

        if attr in GETATTR_PASS and (index := GETATTR_PASS.index(attr)) >= 0:
            return spec if not index else spec.target.__getattribute__(attr)

        if attr in MODULE_SPEC_ATTR_MAP:
            match attr:
                case "__file__":
                    if spec.has_location:
                        return spec.origin
                case "__cached__":
                    if spec.has_location and spec.cached is not None:
                        return spec.cached
                case "__path__":
                    if spec.submodule_search_locations is not None:
                        return spec.submodule_search_locations
                case _:
                    return getattr(spec, MODULE_SPEC_ATTR_MAP[attr])

            return spec.target.__getattribute__(attr)

        # if not conf.NO_DICT_LAZY_ATTR:
        #     module_dict = spec.target.__getattribute__("__dict__")
        #     if attr == "__dict__":
        #         return module_dict
        #     if attr in module_dict:
        #         return module_dict[attr]

        if force := (spec.loader_state.value <= spec.loader.State.LAZY.value):
            assert None is debug.trace(f"[{id(self)}] <get> {spec.loader_state} {spec.name}[.{attr}]")
            spec.loader.exec_module(self, spec, force)

        valu = spec.target.__getattribute__(attr)

        # NOTE: If NO_DICT_LAZY_ATTR is False, we need to make sure that the result of getattribute
        #       is loaded, if it's a module.
        #       Otherwise, we can just trigger the exec and then return it (or raise AttributeError).
        #
        # if conf.NO_DICT_LAZY_ATTR is False and isinstance(valu, ModuleType):
        #     if isinstance(valu, ModuleType):
        #         loader = valu.__loader__
        #         valu.__loader__.exec_module(valu, None, force) if isinstance(loader, type(spec.loader)) else \
        #             valu.__loader__.exec_module(valu)

        return valu

    def __setattr__(self, attr, valu):
        spec = super().__getattribute__("__spec__")

        if attr in SETATTR_PASS:
            if attr == "__spec__":
                return super().__setattr__(attr, valu)

            return spec.target.__setattr__(attr, valu)

        if force := (spec.loader_state.value <= spec.loader.State.LAZY.value):
            assert None is debug.trace(f"[{id(self)}] <set> {spec.loader_state} {spec.name}[.{attr}] [{id(valu)}]")
            spec.loader.exec_module(self, spec, force)

        # NOTE: If NO_DICT_LAZY_ATTR is False, we need to make sure that the target attribute
        #       is loaded, if it's a module attribute.
        #       Otherwise, we can just trigger the exec and then set it.
        #
        # if conf.NO_DICT_LAZY_ATTR is False and isinstance(valu, ModuleType):
        #     loader = valu.__loader__
        #     valu.__loader__.exec_module(valu, None, force) if isinstance(loader, type(spec.loader)) else \
        #         valu.__loader__.exec_module(valu)

        return spec.target.__setattr__(attr, valu)
